
#include "ldb.h"
#include "ldb_iterator.h"
#include <leveldb/env.h>

//#include <iostream>

#define CHECK_DB_IS_OPEN \
if (this->leveldb == NULL) {\
    Nan::ThrowError("[Ldb error] database not open");\
    return leveldb::Status::NotFound("Ldb database not open");\
}\

#define ERROR(MSG) \
Exception::Error(Nan::New(MSG).ToLocalChecked() )

#define PROMIES_ERROR(MSG) \
Local<Value> err = ERROR(MSG);\
promise->Reject(err);\
Nan::ThrowError(err)

namespace Ldb {
    
    Ldb::~Ldb() {
        this->Close();
        if (_options != NULL) {
            HI_DELETE(_options->block_cache);
            HI_DELETE(_options->filter_policy);
            delete _options;
            _options = NULL;
        }
        HI_DELETE(_event);
    }
    
    void Ldb::Close() {
        HI_DELETE(leveldb);
        EmitEvent("close");
    }
    
    bool Ldb::IsOpen() {
        return this->leveldb != NULL;
    }
        
    leveldb::Status Ldb::Open() {
        if (IsOpen()){
            Nan::ThrowError("[Ldb.open error] The database is already open");
            return leveldb::Status::InvalidArgument("[Ldb.open error] The database is already open");
        }
        if (_location.empty()){
            Nan::ThrowError("[Ldb.open error] the database name can't is empty");
            return leveldb::Status::InvalidArgument("[Ldb.open error] the database name can't is empty");
        }
        if (_options == NULL) {
            _options = TranslationOptions(Local<Object>(), _options);
        }
        bool exist_already = IsExist(_location);
        leveldb::Status status = leveldb::DB::Open(*_options, _location, &leveldb);
        if (status.ok()){
            if (!exist_already) {
                EmitEvent("create", "succeed");
            }
            EmitEvent("ready", "succeed");
        }else{
            Nan::ThrowError(status.ToString().c_str());
        }
        return status;
    }
    
    leveldb::Status Ldb::Put(const string& key, const string& value) {
        CHECK_DB_IS_OPEN;
        leveldb::Status status = this->leveldb->Put(_writeOpt, key, value);
        if (status.ok()) {
            EmitEvent("put", "succeed", key, value);
        }else{
            EmitError("put", status, key, value);
        }
        return status;
    }
    
    leveldb::Status Ldb::Get(const string& key, string& value) {
        CHECK_DB_IS_OPEN
        leveldb::Status status = this->leveldb->Get(_readOpt, key, &value);
        if (!status.ok()) {
            EmitError("get", status, key);
        }
        return status;
    }
    
    leveldb::Status Ldb::Del(const string& key) {
        CHECK_DB_IS_OPEN;
        leveldb::Status status = this->leveldb->Delete(_writeOpt, key);
        if (status.ok()) {
            EmitEvent("del", "succeed", key);
        }else{
            EmitError("del", status, key);
        }
        return status;
    }
    
    leveldb::Status Ldb::PutBatch (Batch* batch) {
        CHECK_DB_IS_OPEN;
        leveldb::WriteBatch *leveldb_batch = batch->Export();
        leveldb::Status status = this->leveldb->Write(_writeOpt, leveldb_batch);
        if (status.ok()) {
            if(_event->HasListener("changed")) {
                for(list<Data>::iterator it = batch->datas.begin(); it != batch->datas.end(); it++){
                    EmitEvent(it->type == Data::Del ? "del" : "put", "succeed", it->key, it->value);
                }
            }
        }else{
            EmitError("batch", status);
        }
        delete leveldb_batch;
        return status;
    }

    leveldb::Status Ldb::PutBatch (const Local<Value>& value) {
        Batch* batch = new Batch();
        batch->Add(value);
        leveldb::Status status = this->PutBatch(batch);
        delete batch;
        return status;
    }
    
    bool Ldb::SetOptions(HI_ARGS_INFO, const bool check_error) {
        HI_SCOPE;
        string location = "";
        Local<Object> opt_obj;
        if ( info[0]->IsString() ) {
            location = HI_STR(info[0]);
            if (info.Length() > 1 && info[1]->IsObject()){
                opt_obj = info[1].As<Object>();
            }
        }else if (info[0]->IsObject()) {
            opt_obj = info[0].As<Object>();
            location = HI::Obj_Get(opt_obj, "location", "");
        }
        
        if (!opt_obj.IsEmpty()) {
            _event->On("ready", opt_obj->Get(HI_STRING("onReady")));
            _event->On("create", opt_obj->Get(HI_STRING("onCreate")));
            _event->On("changed", opt_obj->Get(HI_STRING("onChanged")));
            _event->On("error", opt_obj->Get(HI_STRING("onError")));
            _event->On("closed", opt_obj->Get(HI_STRING("onClosed")));
            _readOpt.fill_cache = HI::Obj_Get(opt_obj, "readFillCache", true);
            _writeOpt.sync = HI::Obj_Get(opt_obj, "writeSync", false);
            _options = TranslationOptions(opt_obj, this->_options);
        }
        if (location.empty()) {
            if (check_error) {
                Nan::ThrowError("[Ldb.open error] the database name can't is empty");
            }
            return false;
        }
        this->_location = location;
        return true;
    }
    
    void Ldb::EmitError(const string& from, leveldb::Status& status, const string& key, const string& value) {
        if (_event->HasListener("error")) {
            Local<Object> err = Exception::Error( HI_STRING( status.ToString() ) ).As<Object>();
            err->Set(HI_STRING("type"), HI_STRING(from));
            err->Set(HI_STRING("target"), this->handle());
            if (!key.empty()){
                err->Set(HI_STRING("key"), HI_STRING(key));
            }
            if (!value.empty()){
                err->Set(HI_STRING("value"), HI::Json_parse(value));
            }
            Local<Value> argv[1] = { err };
            _event->Emit("error", 1, argv);
        }
    }
    
    void Ldb::EmitEvent(const string& from, const string& msg, const string& key, const string& value) {
        string type = from;
        if (from == "put" || from == "del"){
            type = "changed";
            this->_state_code ++;
        }else if ( from == "closed" ){
            this->_state_code ++;
        }
        if (_event->HasListener(type)) {
            HI_SCOPE;
            Local<Object> event = Object::New(isolate);
            event->Set(HI_STRING("type"), HI_STRING(from));
            event->Set(HI_STRING("target"), this->handle());
            if (!key.empty()){
                event->Set(HI_STRING("key"), HI_STRING(key));
            }
            if (!value.empty()){
                event->Set(HI_STRING("value"), HI::Json_parse(value));
            }
            if (!msg.empty()) {
                event->Set(HI_STRING("message"), HI_STRING(msg));
            }
            Local<Value> argv[1] = { event };
            _event->Emit(type, 1, argv);
        }
    }
    
    bool Ldb::IsExist(const string& localtion, const leveldb::Options& opt) {
        return opt.env->FileExists(localtion + "/CURRENT");
    }
    
    leveldb::Options* Ldb::TranslationOptions(const Local<Object>& opt_obj, leveldb::Options* opt) {
        bool create = false;
        int32_t block_cache = 0;
        int32_t filter_policy = 0;
        if (opt == NULL) {
            create = true;
            opt = new leveldb::Options();
            opt->create_if_missing      = true;
            opt->error_if_exists        = false;
            opt->compression            = leveldb::kSnappyCompression;
            opt->write_buffer_size      = (4 << 20);
            opt->block_size             = 4096;
            opt->max_open_files         = 1000;
            opt->block_restart_interval = 16;
            opt->max_file_size          = (2 << 20);
            block_cache = (8 << 20);
            filter_policy = 10;
        }
        if (!opt_obj.IsEmpty()){
            opt->create_if_missing      = HI::Obj_Get(opt_obj, "createIfMissing", opt->create_if_missing);
            opt->error_if_exists        = HI::Obj_Get(opt_obj, "errorIfExists", opt->error_if_exists);
            opt->write_buffer_size      = HI::Obj_Get(opt_obj, "writeBufferSize", int32_t(opt->write_buffer_size));
            opt->block_size             = HI::Obj_Get(opt_obj, "blockSize", int32_t(opt->block_size));
            opt->max_open_files         = HI::Obj_Get(opt_obj, "maxOpenFiles", opt->max_open_files);
            opt->block_restart_interval = HI::Obj_Get(opt_obj, "blockRestartInterval", opt->block_restart_interval);
            opt->max_file_size          = HI::Obj_Get(opt_obj, "maxFileSize", int32_t(opt->max_file_size));
            opt->compression            = HI::Obj_Get(opt_obj, "compression", true) ? leveldb::kSnappyCompression : leveldb::kNoCompression;
            block_cache = HI::Obj_Get(opt_obj, "cacheSize", block_cache);
            filter_policy = HI::Obj_Get(opt_obj, "filterPolicy", filter_policy);
        }
        if (create && block_cache > 0) {
            opt->block_cache = leveldb::NewLRUCache( block_cache );
        }
        if (create && filter_policy > 0) {
            opt->filter_policy = leveldb::NewBloomFilterPolicy( filter_policy );
        }
        return opt;
    }
    
    //////////////////////////////////////////////////////////////////////////////
    
    HI_CLASS_INIT(Ldb) {
        _readOpt.fill_cache = true;
        _writeOpt.sync = false;
        _state_code = 0;
        _event = new EventEmitter(this);
        _options = NULL;
        leveldb = NULL;
        if (info.Length() > 0) {
            SetOptions(info);
        }
    }
    
    HI_GETTER(Ldb::jsGetterLocation) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN(HI_STRING(_this->_location));
    }
    
    HI_GETTER(Ldb::jsGetterIsOpen) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN(HI_BOOLEAN(_this->IsOpen()));
    }
    
    HI_GETTER(Ldb::jsGetterStateCode) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN(Number::New(isolate, _this->_state_code));
    }
    
    HI_METHOD(Ldb::jsOpen) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        if (info.Length() > 0 ){
            if (_this->SetOptions(info, true)) {
                _this->Open();
            }
        }else{
            _this->Open();
        }
        HI_RETURN_THIS;
    }
    
    HI_METHOD(Ldb::jsClose) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        _this->Close();
        HI_RETURN_THIS;
    }
    
    HI_METHOD(Ldb::jsPut) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        leveldb::Status status;
        int info_len = info.Length();
        if (info_len > 0) {
            if (info[0]->IsObject()) {
                status = _this->PutBatch(info[0]);
            }else{
                if (info_len < 2) {
                    return Nan::ThrowError("[Ldb.put error] missing argument, refer to: put(key, value)");
                }
                status = _this->Put(HI_TO_STR(info[0]), HI_STR(HI::Json_stringify(info[1])));
            }
        }
        HI_RETURN( HI_BOOLEAN(status.ok()) );
    }
    
    HI_METHOD(Ldb::jsDel) {
        HI_SCOPE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        int info_len = info.Length();
        if (info_len < 1) {
            return Nan::ThrowError("[Ldb.del error] missing argument, refer to: delete(key, sync=false)");
        }
        leveldb::Status status;
        if (info[0]->IsArray()) {
            status = _this->PutBatch(info[0]);
        }else if( HI_INSTANCE_OF_CLASS(info[0], Iterator) ){
            Iterator *iter = HI_UNWRAP(Iterator, info[0].As<Object>());
            while (iter->Read()) {
                status = _this->Del(iter->currentKey);
            }
            iter->Release();
        }else{
            status = _this->Del(HI_TO_STR(info[0]));
        }
        HI_RETURN( HI_BOOLEAN(status.ok()) );
    }
    
    HI_METHOD(Ldb::jsGet) {
        HI_SCOPE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        int info_len = info.Length();
        if (info_len < 1) {
            return Nan::ThrowError("[Ldb.get error] missing argument, refer to: get(key)");
        }
        leveldb::Status status;
        string ref;
        if (info[0]->IsArray()){
            Local<Object> res = Object::New(isolate);
            Local<Array> names =info[0].As<Array>();
            for (unsigned int i = 0; i < names->Length(); i++) {
                string key = HI_TO_STR(names->Get(i));
                if (_this->Get(key, ref).ok()) {
                    res->Set(names->Get(i), HI::Json_parse(ref));
                }
            }
            HI_RETURN(res);
        }else{
            string key = HI_TO_STR(info[0]);
            if (_this->Get(key, ref).ok()) {
                HI_RETURN(HI::Json_parse(ref));
            }
        }
    }
    
    HI_METHOD(Ldb::jsForEach) {
        HI_SCOPE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        int info_len = info.Length();
        if (info_len < 1) {
            return Nan::ThrowError("[Ldb.forEach error] missing argument, refer to: forEach(callback)");
        }
        if (!info[0]->IsFunction()) {
            return Nan::ThrowError("[Ldb.forEach error] requires a callback argument");
        }
        Nan::Callback *callback = new Nan::Callback(Local<Function>::Cast(info[0]));
        Local<Object> iter_obj = Iterator::NewInstance(_this, info[1], callback);
        Iterator* iter = HI_UNWRAP(Iterator, iter_obj);
        while (iter->Read()) {
            if (!iter->Call()){
                break;
            }
        }
        iter->Release();
    }
    
    HI_METHOD(Ldb::jsGetIterator) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN( Iterator::NewInstance(_this, info[0], NULL) );
    }
    
    HI_METHOD(Ldb::jsAsyncPut) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN_PROMISE(promise);
        int info_len = info.Length();
        if (info_len > 0) {
            Batch* batch = new Batch();
            if (info[0]->IsObject()) {
                batch->Add(info[0]);
            }else{
                if (info_len < 2) {
                    PROMIES_ERROR("[Ldb.put error] missing argument, refer to: put(key, value, sync=false)");
                    return;
                }
                batch->Put(HI_TO_STR(info[0]), HI_STR(HI::Json_stringify(info[1])));
            }
            BatchWorker *worker = new BatchWorker(_this, promise, batch);
            Nan::AsyncQueueWorker(worker);
        }
    }
    
    
    HI_METHOD(Ldb::jsAsyncDel) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN_PROMISE(promise);
        int info_len = info.Length();
        if (info_len < 1) {
            PROMIES_ERROR("[Ldb.del error] missing argument, refer to: delete(key, sync=false)");
            return;
        }
        if (HI_INSTANCE_OF_CLASS(info[0], Iterator)) {
            Iterator *iter = HI_UNWRAP(Iterator, info[0].As<Object>());
            IteratoWorker* worker = new IteratoWorker(_this, promise, iter, IteratoWorker::Operate::Delete);
            Nan::AsyncQueueWorker(worker);
            return;
        }
        Batch* batch = new Batch();
        if (info[0]->IsObject()) {
            batch->Add(info[0]);
        }else{
            batch->Del(HI_TO_STR(info[0]));
        }
        BatchWorker *worker = new BatchWorker(_this, promise, batch);
        Nan::AsyncQueueWorker(worker);
    }
    
    HI_METHOD(Ldb::jsAsyncGet) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN_PROMISE(promise);
        int info_len = info.Length();
        if (info_len < 1) {
            PROMIES_ERROR("[Ldb.get error] missing argument, refer to: get(key)");
            return;
        }
        if (info[0]->IsArray()){
            Local<Array> names =info[0].As<Array>();
            list<string>* keys = new list<string>();
            for (unsigned int i = 0; i < names->Length(); i++) {
                keys->push_back(HI_TO_STR(names->Get(i)));
            }
            GetListWorker* worker = new GetListWorker(_this, promise, keys);
            Nan::AsyncQueueWorker(worker);

        }else{
            string key = HI_TO_STR(info[0]);
            GetWorker* worker = new GetWorker(_this, promise, key);
            Nan::AsyncQueueWorker(worker);
        }
    }
    
    HI_METHOD(Ldb::jsAsyncForEach) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Ldb);
        HI_RETURN_PROMISE(promise);
        int info_len = info.Length();
        if (info_len < 1) {
            PROMIES_ERROR("[Ldb.forEach error] missing argument, refer to: forEach(callback)");
            return;
        }
        if (!info[0]->IsFunction()) {
            PROMIES_ERROR("[Ldb.forEach error] requires a callback argument");
            return;
        }
        Nan::Callback *callback = new Nan::Callback(Local<Function>::Cast(info[0]));
        Local<Object> iter_obj = Iterator::NewInstance(_this, info[1], callback);
        Iterator *iter = HI_UNWRAP(Iterator, iter_obj);
        ForEachWorker* worker = new ForEachWorker(_this, promise, iter);
        Nan::AsyncQueueWorker(worker);
    }
    
    HI_METHOD(Ldb::jsAddListener) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        _this->_event->On(HI_STR(info[0]), info[1], HI_BOOL(info[2], false));
        HI_RETURN_THIS;
    }
    
    HI_METHOD(Ldb::jsRemoveListener) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        _this->_event->Un(HI_STR(info[0]), Local<Function>::Cast(info[1]));
        HI_RETURN_THIS;
    }
    
    HI_METHOD(Ldb::jsMatch) {
        HI_OBJECT_UNWRAP_THIS(Ldb);
        if(info.Length() >= 1) {
            string prefix = HI_TO_STR(info[0]);
            int32_t offset = info[1]->IsNumber() ? info[1]->Int32Value() : 0;
    
            leveldb::ReadOptions opt;
            leveldb::Iterator *iter = _this->leveldb->NewIterator(opt);
            if (offset < 0) {
                iter->Seek( Iterator::NextKey(prefix) );
                if (!iter->Valid()) {
                    iter->SeekToLast();
                    if (iter->Valid()) {
                        if(prefix.compare(iter->key().ToString().c_str()) > 0){
                            delete iter;
                            return;
                        }
                        offset ++;
                    }
                }
                while (iter->Valid() && offset ++ < 0) {
                    iter->Prev();
                }
            }else{
                iter->Seek(prefix);
                while (iter->Valid() &&offset -- > 0) {
                    iter->Next();
                }
            }
            if (iter->Valid()) {
                string key = iter->key().ToString();
                if(prefix.compare(key.c_str()) <= 0){
                    HI_SCOPE;
                    Local<Object> res = Object::New(isolate);
                    res->Set(HI_STRING("key"), HI_STRING(key));
                    res->Set(HI_STRING("value"), HI::Json_parse(iter->value().ToString()));
                    HI_RETURN_VALUE(res);
                }
            }
            delete iter;
        }
    }
    
    HI_CLASS_DEFINE_TEMPLATE(Ldb) {
        HI::SetPrototype(tpl, "location", jsGetterLocation);
        HI::SetPrototype(tpl, "isOpen", jsGetterIsOpen);
        HI::SetPrototype(tpl, "stateCode", jsGetterStateCode);
        HI::SetPrototype(tpl, "open", jsOpen);
        HI::SetPrototype(tpl, "close", jsClose);
        HI::SetPrototype(tpl, "put", jsPut);
        HI::SetPrototype(tpl, "get", jsGet);
        HI::SetPrototype(tpl, "del", jsDel);
        HI::SetPrototype(tpl, "forEach", jsForEach);
        HI::SetPrototype(tpl, "getIterator", jsGetIterator);
        HI::SetPrototype(tpl, "asyncPut", jsAsyncPut);
        HI::SetPrototype(tpl, "asyncGet", jsAsyncGet);
        HI::SetPrototype(tpl, "asyncDel", jsAsyncDel);
        HI::SetPrototype(tpl, "asyncForEach", jsAsyncForEach);
        HI::SetPrototype(tpl, "addListener", jsAddListener);
        HI::SetPrototype(tpl, "removeListener", jsRemoveListener);
        
        HI::SetPrototype(tpl, "match", jsMatch);
    }
    
}
