
#include "ldb_async.h"

namespace Ldb {
    
    PromiseWorker::PromiseWorker(Ldb* ldb, PromiseObj promise): Nan::AsyncWorker(NULL), ldb(ldb) {
        SaveToPersistent("promise", promise);
    }
    
    PromiseWorker::PromiseWorker(Ldb* ldb, PromiseObj promise, Nan::Callback* callback): Nan::AsyncWorker(callback), ldb(ldb) {
        SaveToPersistent("promise", promise);
    }
    
    Local<Promise::Resolver> PromiseWorker::GetPromise() {
        Local<Value> v = GetFromPersistent("promise");
        if (v->IsPromise()) {
            return v.As<Promise::Resolver>();
        }
        return Local<Promise::Resolver>();
    }
    
    void PromiseWorker::HandleOKCallback() {
        Local<Promise::Resolver> promise = GetPromise();
        if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
            if (status.ok()) {
                promise->Resolve( Nan::Undefined() );
            }else{
                promise->Reject( Exception::Error(HI_STRING(status.ToString())));
            }
        }
    }
    
    void PromiseWorker::HandleErrorCallback() {
        Local<Promise::Resolver> promise = GetPromise();
        if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
            promise->Reject(Exception::Error(Nan::New(ErrorMessage()).ToLocalChecked() ));
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    
    GetWorker::GetWorker (Ldb* ldb, PromiseObj promise, const string& key): PromiseWorker(ldb, promise) {
        _key = key;
    }
    
    void GetWorker::Execute() {
        if (ldb != NULL){
            status = ldb->Get(_key, _result);
        }
    }
    
    void GetWorker::HandleOKCallback(){
        Local<Promise::Resolver> promise = GetPromise();
        if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
            HI_SCOPE;
            if (status.ok()) {
                if (!_result.empty()) {
                    promise->Resolve( HI::Json_parse(_result) );
                }else{
                    promise->Resolve(HI_UNDEFINED);
                }
            }else{
                if (status.IsNotFound()){
                    promise->Resolve(HI_UNDEFINED);
                }else{
                    promise->Reject(Exception::Error(Nan::New(status.ToString()).ToLocalChecked()));
                }
            }
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    
    GetListWorker::GetListWorker (Ldb* ldb, PromiseObj promise, list<string>* keys): PromiseWorker(ldb, promise){
        _keys = keys;
    }
    
    GetListWorker::~GetListWorker (){
        HI_DELETE(_keys);
    }
    
    void GetListWorker::Execute() {
        if (ldb != NULL && _keys != NULL){
            for(list<string>::iterator it = _keys->begin(); it != _keys->end(); it++){
                string value;
                leveldb::Status status = ldb->Get(*it, value);
                if (status.ok()) {
                    _results.push_back( Data(Data::Cache, *it, value) );
                }else{
                    _keys->erase(it);
                }
            }
        }
    }
    
    void GetListWorker::HandleOKCallback(){
        Local<Promise::Resolver> promise = GetPromise();
        if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
            HI_SCOPE;
            if (!_results.empty()) {
                Local<Object> res = Object::New(isolate);
                for(list<Data>::iterator it = _results.begin(); it != _results.end(); it++){
                    res->Set(HI_STRING(it->key), HI::Json_parse(it->value));
                }
                promise->Resolve( res );
            }else{
                promise->Resolve( HI_UNDEFINED );
            }
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    
    BatchWorker::BatchWorker (Ldb* ldb, PromiseObj promise, Batch* batch): PromiseWorker(ldb, promise) {
        _batch = batch;
    };
    
    BatchWorker::~BatchWorker() {
        HI_DELETE(_batch);
    }
    
    void BatchWorker::Execute() {
        if (ldb != NULL){
            status = ldb->PutBatch(_batch);
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    
    ForEachWorker::ForEachWorker (Ldb* ldb, PromiseObj promise, Iterator *iter): Nan::AsyncWorker(NULL) {
        _iter = iter;
        SaveToPersistent("promise", promise);
        SaveToPersistent("iterator", iter->handle());
    };
    
    ForEachWorker::ForEachWorker (ForEachWorker* work): Nan::AsyncWorker(NULL) {
        _iter = work->_iter;
        SaveToPersistent("promise", work->GetFromPersistent("promise"));
        SaveToPersistent("iterator", _iter->handle());
    };
    
    void ForEachWorker::Release() {
        if (_iter != NULL) {
            _iter->Release();
            _iter = NULL;
        }
        
        Local<Value> v = GetFromPersistent("promise");
        if (v->IsPromise()) {
            Local<Promise::Resolver> promise = v.As<Promise::Resolver>();
            if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
                promise->Resolve(Nan::Undefined());
            }
        }
    }
    
    void ForEachWorker::Execute() {
        if (!_iter->Read()){
            Release();
        }
    }
    
    void ForEachWorker::HandleOKCallback(){
        if (_iter == NULL) {
            Release();
            return;
        }
        if (!_iter->Call() || _iter->IsEnd()){
            Release();
            return;
        }
        Nan::AsyncQueueWorker(new ForEachWorker(this));
    }
    
    void ForEachWorker::HandleErrorCallback() {
        Release();
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////
    
    IteratoWorker::IteratoWorker (Ldb* ldb, PromiseObj promise, Iterator *iter, Operate operate): PromiseWorker(ldb, promise) {
        _operate = operate;
        _iter = iter;
        SaveToPersistent("iterator", iter->handle());
    }
    
    void IteratoWorker::Execute () {
        if (_operate == Read) {
            _iter->Read();
        }else if (_operate == Delete) {
            while (_iter->Read()) {
                status = ldb->Del(_iter->currentKey);
                if (!status.ok()) {
                    break;
                }
            }
        }
    }
    
    void IteratoWorker::HandleOKCallback() {
        Local<Promise::Resolver> promise = GetPromise();
        if ( !promise.IsEmpty() && promise->GetPromise()->State() == Promise::PromiseState::kPending ) {
            HI_SCOPE;
            if (_operate == Read) {
                promise->Resolve( _iter->handle() );
            }else if (_operate == Delete) {
                promise->Resolve( HI_BOOLEAN(status.ok()) );
            }
        }
    }
    
}

