
#include "ldb_iterator.h"
#include "ldb.h"

//#include <iostream>

namespace Ldb {
    
    HI_CLASS_INIT(Iterator) {
        ldb = NULL;
        _callback = NULL;
        _iterator = NULL;
        _readOpt = NULL;
        _status = 0;
        _opt_keyStart = "";
        _opt_keyEnd = "";
        _opt_prefix = "";
        _opt_prefixEnd = "";
        _opt_limit = -1;
        _opt_reverse = false;
        _opt_snapshot = true;
    }
    
    Local<Object> Iterator::NewInstance(Ldb* ldb, const Local<Value>& opt, Nan::Callback *callback){
        HI_ISOLATE;
        Local<Object> iter_obj = HI_CLASS_NEW_INSTANCE(Iterator, 0, NULL);
        Iterator *iter = HI_UNWRAP(Iterator, iter_obj);
        iter->ldb = ldb;
        iter->_opt_limit = -1;
        iter->_callback = callback;
        
        if (!opt.IsEmpty() && opt->IsObject()) {
            Local<Object> opt_obj = opt.As<Object>();
            iter->_opt_keyStart = HI::Obj_Get(opt_obj, "start", "");
            iter->_opt_keyEnd = HI::Obj_Get(opt_obj, "end", "");
            iter->_opt_prefix = HI::Obj_Get(opt_obj, "prefix", "");
            iter->_opt_limit = HI::Obj_Get(opt_obj, "limit", -1);
            iter->_opt_reverse = HI::Obj_Get(opt_obj, "reverse", false);
            iter->_opt_snapshot = HI::Obj_Get(opt_obj, "snapshot", true);
        }
        HI::SetProperty(iter_obj, "target", ldb->handle(), true, true, false);
        return iter_obj;
    }

    Iterator::~Iterator() {
        this->Release();
    }
    
    void Iterator::Release() {
        if (_status == 9) {
            return;
        }
        _status = 9;
        // currentKey = "";
        // currentValue = "";
        if (_readOpt != NULL){
            if (_readOpt->snapshot != NULL && ldb != NULL && ldb->leveldb != NULL) {
                ldb->leveldb->ReleaseSnapshot(_readOpt->snapshot);
            }
            HI_DELETE(_readOpt);
        }
        HI_DELETE(_iterator);
        HI_DELETE(_callback);
        Local<Object> _this_obj = this->handle();
        if (!_this_obj.IsEmpty() && _this_obj->Has(HI_STRING("target"))) {
            _this_obj->Delete(HI_STRING("target"));
        }
    }
    
    bool Iterator::IsInited() {
        return _status > 0;
    }
    
    bool Iterator::IsEnd() {
        return _status >= 8;
    }
    
    bool Iterator::IsValid() {
        if (_opt_limit == 0 || _iterator == NULL || !_iterator->Valid()) {
            return false;
        }
        bool hasPrefixEnd = !_opt_prefixEnd.empty();
        if (_opt_reverse) {
            bool hasStart = !_opt_keyStart.empty();
            if (hasPrefixEnd || hasStart) {
                const char* key = this->_iterator->key().data();
                if (hasPrefixEnd) {
                    if(_opt_prefixEnd.compare(key) > 0){
                        return false;
                    }
                }
                if (hasStart) {
                    if(_opt_keyStart.compare(key) >= 0){
                        return false;
                    }
                }
            }
        }else{
            bool hasEnd = !_opt_keyEnd.empty();
            if (hasPrefixEnd || hasEnd) {
                const char* key = this->_iterator->key().data();
                if (hasPrefixEnd) {
                    if(_opt_prefixEnd.compare(key) <= 0){
                        return false;
                    }
                }
                if (hasEnd) {
                    if(_opt_keyEnd.compare(key) <= 0){
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    void Iterator::Init() {
        if (IsInited() || ldb == NULL || ldb->leveldb == NULL) {
            return;
        }
        _status = 1;
        currentKey = "";
        currentValue = "";
        _readOpt = new leveldb::ReadOptions();
        if (_opt_snapshot) {
            _readOpt->snapshot = this->ldb->leveldb->GetSnapshot();
        }
        _readOpt->fill_cache = true;
        _iterator = ldb->leveldb->NewIterator(*_readOpt);
    
        bool hasStart = !_opt_keyStart.empty();
        bool hasPrefix = !_opt_prefix.empty();
        bool hasEnd = !_opt_keyEnd.empty();
        
        if (_opt_reverse) {
            if (hasEnd) {
                this->_iterator->Seek(_opt_keyEnd);
                if (this->_iterator->Valid()){
                    this->_iterator->Prev();
                }else{
                    this->_iterator->SeekToLast();
                    if (this->_iterator->Valid()) {
                        if(_opt_keyEnd.compare( this->_iterator->key().ToString().c_str() ) > 0){
                            this->Release();
                            return;
                        }
                    }
                }
            }else if (hasPrefix){
                this->_iterator->Seek( NextKey(_opt_prefix) );
                if (this->_iterator->Valid()){
                    this->_iterator->Prev();
                }else{
                    this->_iterator->SeekToLast();
                    if (this->_iterator->Valid()) {
                        if(_opt_prefix.compare( this->_iterator->key().ToString().c_str() ) > 0){
                            this->Release();
                            return;
                        }
                    }
                }
            }else{
                this->_iterator->SeekToLast();
            }
            
            if(hasPrefix) {
                _opt_prefixEnd = _opt_prefix;
            }else if (hasStart) {
                
            }
        }else{
            if (hasStart) {
                this->_iterator->Seek(_opt_keyStart);
            }else if (hasPrefix){
                _iterator->Seek(_opt_prefix);
            }else{
                this->_iterator->SeekToFirst();
            }
            if (hasPrefix) {
                _opt_prefixEnd = NextKey(_opt_prefix);
            }else if (hasEnd) {
                
            }
        }

    }
    
    bool Iterator::Read() {
        if (_status == 0){
            Init();
            if (IsValid()){
                return Read();
            }
            return false;
        }
        if (_iterator == NULL){
            this->Release();
            return false;
        }
        if (IsEnd()) {
            return false;
        }
        if (!_iterator->Valid()){
            _status = 8;
            return false;
        }
        currentKey = _iterator->key().ToString();
        currentValue = _iterator->value().ToString();
        
        if(_opt_reverse){
            _iterator->Prev();
        }else{
            _iterator->Next();
        }
        _opt_limit --;
        _status = !IsValid() ? 8 : _status;
        return true;
    }
    
    bool Iterator::Call() {
        if (_callback != NULL){
            Local<v8::Value> args[] = { this->handle() };
            Local<v8::Value> ref = _callback->Call(1, args);
            if (IsEnd() || (ref->IsBoolean() && ref->ToBoolean() == Nan::False())) {
                return false;
            }
            return true;
        }
        return false;
    }
    
    string Iterator::NextKey(const string& key){
        auto s = key.size();
        return key.substr(0, s-1) + char(key[s-1]+1);
    }
    
    string Iterator::PrevKey(const string& key){
        auto s = key.size();
        return key.substr(0, s-1) + char(key[s-1]-1);
    }
    
    /////////////////////////////////////////////////////////////
    
    HI_GETTER(Iterator::jsGetterKey) {
        HI_OBJECT_UNWRAP_THIS(Iterator);
        if (!_this->currentKey.empty()) {
            HI_RETURN( HI_STRING(_this->currentKey) );
        }
    }
    
    HI_GETTER(Iterator::jsGetterValue) {
        HI_OBJECT_UNWRAP_THIS(Iterator);
        if (!_this->currentValue.empty()) {
            HI_RETURN( HI::Json_parse(_this->currentValue) );
        }
    }
    
    HI_GETTER(Iterator::jsGetterText) {
        HI_OBJECT_UNWRAP_THIS(Iterator);
        if (!_this->currentValue.empty()) {
            HI_RETURN( HI_STRING(_this->currentValue) );
        }
    }
    
    HI_GETTER(Iterator::jsGetterIsEnd) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Iterator);
        HI_RETURN( HI_BOOLEAN(_this->IsEnd()) );
    }
    
    HI_METHOD(Iterator::jsRead) {
        HI_OBJECT_UNWRAP_THIS(Iterator);
        if (_this->Read()) {
            HI_RETURN_THIS;
        }else {
            _this->Release();
        }
    }
    
    HI_METHOD(Iterator::jsAsyncRead) {
        HI_ISOLATE;
        HI_OBJECT_UNWRAP_THIS(Iterator);
        HI_RETURN_PROMISE(promise);
        IteratoWorker* worker = new IteratoWorker(_this->ldb, promise, _this, IteratoWorker::Operate::Read);
        Nan::AsyncQueueWorker(worker);
    }
    
    HI_METHOD(Iterator::jsBreak) {
        HI_OBJECT_UNWRAP_THIS(Iterator);
        _this->Release();
    }
    
    HI_CLASS_DEFINE_TEMPLATE(Iterator) {
        HI::SetPrototype(tpl, "key", jsGetterKey, 0, true, false, true);
        HI::SetPrototype(tpl, "value", jsGetterValue, 0, true, false, true);
        HI::SetPrototype(tpl, "text", jsGetterText, 0, true, true, true);
        HI::SetPrototype(tpl, "isEnd", jsGetterIsEnd, 0, true, false, true);
        HI::SetPrototype(tpl, "read", jsRead, true, true, true);
        HI::SetPrototype(tpl, "asyncRead", jsAsyncRead, true, true, true);
        HI::SetPrototype(tpl, "break", jsBreak, true, true, true);
    }
    
}

